 
// File ML1_MDS
 
// Machine-dependent logic for ML/I macro processor on EMAS 2900

// BCPL version by R.D. Eager   University of Kent	MCMLXXXII
//       hacked by Dave Cannon	to run under Windows	MMXII
 
// Version: E3.1   (also alter MANIFESTs below)


// SECTION "ML1"


/* Use of S-variables
   ------------------

S1-S9 are used by the machine-independent logic
-----------------------------------------------

  S1   -  0 => Don't insert startlines on input
          1 => Insert startlines on input
  S2   -  Source text line number
  S3   -  0 => Give error message if warning marker not followed by macro name
          1 => Don't give error message if warning marker not followed by macro name
  S4   -  0 => Don't suppress context print-out on MC-NOTE call
          1 => Suppress context print-out on MC-NOTE call
  S5   -  Count of processing errors
  S6   -  Not used
  S7   -  Not used
  S8   -  Not used
  S9   -  Not used

S10 upwards are used by the machine-dependent logic
---------------------------------------------------

  S10  -  Input switch:
          0 => end of process
          1 -> maxinputs selects stream
          101 -> 100 + maxinputs selects and rewinds stream
  S11  -  Not used
  S12  -  Remaining line quota on debugging file
  S13  -  Not used
  S14  -  Not used
  S15  -  Not used
  S16  -  Character to be translated on input
  S17  -  Character to replace S16
  S18  -  0 => Don't give environment printout at end of process
          1 => Give environment printout at end of process
  S19  -  Line number of output text
  S20  -  0 => Don't list output text
          1 => List output text without line numbers
          2 => List output text with line numbers
  S21  -  0 => Don't send output to primary output
          1 => Send output to primary output
  S22  -  0 => Don't send output to secondary output
          1 => Send output to secondary output
  S23  -  Reverted input switch
*/

MANIFEST $(
					//	Major version number
version    = 3
					//	Edit number within major version
edit       = 1
//			 1			Dave Cannon's initial Windows hack:	27-Nov-2012
//			 0			Bob Eager's original EMAS 2900 implementation
$)

GET "ML1HDR"
 
STATIC $(
journal          = 0
error_width      = 0   // Line length on debugging file
error_ch_count   = 0   // Count of characters on current line of debugging file
nlpending_       = 0   // TRUE iff last input character was a newline
input_streams    = 0   // Vector of input stream pointers
listing_output   = 0   // Stream pointer for listing file
primary_output   = 0   // Stream pointer for primary output file
secondary_output = 0   // Stream pointer for secondary output file
error_output     = 0   // Stream pointer for debugging file
$)

MANIFEST $(
maxinputs = 3                           // Maximum number of input files
$)

LET start(pvec, argc) BE
/* Main initialisation routine.  Reads and decodes the parameter string,
opens files, and obtains workspace.  */
$( LET is = VEC maxinputs
   AND os = VEC 4
   AND workad = 0
   AND flag = 0
   AND res = 0
   
   AND argv = pvec
   AND io_strms = 0
   AND ix = 0

   journal := output()
   error_output := journal
   selectoutput(error_output)

   /* Set up input streams from parameter string */

   input_streams := is
   FOR i = 0 TO maxinputs DO input_streams!i := 0
                                        // 'input_streams!0' must be zero
/*
   FOR i = 1 TO maxinputs DO
   $( LET name = VEC maxstrlength/bytesperword
      LET term = get_name(param, name)
      AND strp = 0

      IF getbyte(name, 0) NE 0 THEN
      $( strp := findinput(name)
         IF strp < 0 THEN fail(strp, name)
         input_streams!i := strp
      $)
      IF term = '/' LOGOR term = endstreamch BREAK
      IF i = maxinputs & getbyte(param, 0) NE 0 THEN fail(277, 0)
                                        // Too many input files
   $)
*/
   IF input_streams!1 = 0 THEN input_streams!1 := sysin
                                        // So that there is at least one input stream

   /* Set up output streams from parameter string */
/*
   FOR i = 1 TO 4 DO os!i := 0

   IF getbyte(param, 0) NE 0 THEN
   $( FOR i = 1 TO 4 DO
      $( LET name = VEC maxstrlength/bytesperword
         LET term = get_name(param, name)
         AND strp = 0

         IF getbyte(name, 0) NE 0 THEN
         $( strp := findoutput(name)
            IF strp < 0 THEN fail(strp, name)
            os!i := strp
         $)
         IF term = endstreamch BREAK
         IF i = 4 & getbyte(param, 0) NE 0 THEN fail(215, 0)
                                        // Too many parameters
      $)
   $)
*/

   io_strms := input_streams
   ix := 1
   
   FOR pct = 0 TO argc-1 DO
   $( LET strp = 0
      IF getbyte(argv!pct, 1) = '/' THEN 
      $( io_strms := os
	     ix := 1
		 LOOP
	  $)
	  TEST io_strms = input_streams THEN $(
	    IF ix > maxinputs THEN fail(277, 0)
                                        // Too many input files
		strp := findinput(argv!pct)
//		writef("strp=%N on %S*n", strp, argv!pct)
		$) ELSE $(
		IF ix > 4 THEN fail(215, 0)
                                        // Too many parameters
		strp := findoutput(argv!pct)
		$)
	  IF strp = 0 THEN fail(199, argv!pct)
	  io_strms!ix := strp
	  ix := ix+1
   $)

   primary_output    := os!1
   secondary_output  := os!2
   listing_output    := os!3
   IF os!4 NE 0 THEN error_output := os!4
   IF primary_output = 0 THEN primary_output := sysout

   setwork(@workad, @flag)              // Set up work area
// writef("setwork(%N, %N) returns...*n", workad, flag)
   IF flag NE 0 THEN fail(flag, "Unable to allocate workspace")
// workad := workad >> 2                // Convert to word address.  DC: not required - setwork() now returns a word pointer

   writef("ML/I macro processor - version E%N.%N*N", version, edit)
   prompt("ML/I: ")

// error_width := b_uinfi(2) = 2 -> 132, b_uinfi(15)
                                        // Width of printer/terminal output
   error_width := 132					// DC: Fix width of output to value
   error_ch_count := 0

   /* Pass control to the machine-independent logic */

// writef("...about to call milogic(%N, %X8)*n", workad, workad!2)
   res := milogic(workad, workad!2)
                                        // Address and size, in words

   /*  'res'  is  set  the  number  of  ML/I processing errors.  On this
   implementation, we just set RETURN CODE to 1000  if  there  were  any
   errors at all, otherwise it is set to zero (meaning "success").  */

   stop(res = 0 -> 0, 1000)
$)

AND get_name(s, name) = VALOF
/* Returns the next filename from the string 's', in 'name'.  Yields the
character  that  terminated  the  name, which may be ',' or '/'.  Yields
'endstreamch' if 's' is exhausted.  'name' and its terminator  (if  any)
are removed from 's'.  */
$( LET len = 0
   AND term = endstreamch

   FOR i = 1 TO getbyte(s, 0) DO
   $( LET c= getbyte(s, i)

      IF c = ',' LOGOR c = '/' THEN
      $( term := c
         BREAK
      $)
      len := len + 1
      putbyte(name, len, c)
   $)
   putbyte(name, 0, len)

   len := len + 1
   FOR i = 1 TO getbyte(s, 0) - len DO putbyte(s, i, getbyte(s, (i + len)))
   $( LET l = getbyte(s, 0) - len
      IF l < 0 THEN l := 0
      putbyte(s, 0, l)
   $)

   RESULTIS term
$)

/*
AND b_uinfi(n) = VALOF
// Interface to the EMAS function UINFI, which  is  used  for  obtaining
// information about the environment in which ML/I is running.
$( EXTERNAL $( uinfi $)
   RESULTIS uinfi(n)
$)
*/
  
AND setwork(ad, flag) BE
/*  Interface  to  the EMAS system routine SETWORK, which is used to (if
necessary) create and initialise the standard workfile, T#WRK.  */
$( // EXTERNAL $( s_setwork : "S#SETWORK" $)
   MANIFEST $( wksize = #X1000000 $)
   LET a, f = 0, 0

// s_setwork(#X28000001, ad << 2, #X28000001, flag << 2)	// DC: replaced by...
   !flag := 0								// Assume success (glass half full!)
   !ad := getvec(wksize)
   IF !ad = 0 DO $(
      !flag := 999
	  RETURN
	  $)
   a := !ad
   a!0 := wksize
   a!1 := !ad
   a!2 := wksize
   a!3 := !flag
// writef("getvec() returns with %N, flag=%N*n", !ad, !flag)
$)

AND fail(flag, name) BE
/* Routine called on a fatal error concerned with the system  interface,
or  a parameter error.  Prints a suitable message on .OUT, sets the EMAS
return code, and exits.  */
$( // flag := ABS flag
   selectoutput(journal)
   writef("*NML1 fails: %S", ssmessage(flag, name))
   stop(flag)
$)

AND mderpr(format, a, b, c, d, e, f, g, h, i, j, k) BE
/* Standard ML/I machine-dependent routine; used to  output  information
to  the debugging file.  Handles wrap-round of long lines of output, and
enforces the lines limit controlled by the value of S12.  */
$( STATIC $( save_wrch = 0 $)

   LET t = @a
   AND o = output()

   LET wrid() BE
      FOR i = 0 TO idlen - 1 DO
      $( LET c = idpt!i

         TEST c = slrep THEN
            writes("(SL)")
         OR
            wrch(c)
      $)

   AND error_wrch(c) BE
   $( TEST c = '*N' THEN
      $( error_ch_count := 0
         !at_s12 := !at_s12 - 1
         IF !at_s12 < 0 THEN
         $( !at_s12 := maxint           // To allow environment dump, etc.
            wrch := save_wrch
            newline()
            prerr()
            mderpr("*NDebugging file lines quota exhausted*N")
            mdabort()
         $)
      $)
      OR
      $( IF error_ch_count = error_width THEN
         $( wrch('*N')
            error_ch_count := 0
         $)
         error_ch_count := error_ch_count + 1
      $)
      save_wrch(c)
   $)
 
   save_wrch := wrch
   wrch := error_wrch

   selectoutput(error_output)

   FOR p = 1 TO getbyte(format, 0) DO
   $( LET k = getbyte(format, p)
 
      TEST k = '%' THEN
      $( LET q, n = !t, 0
         AND type = getbyte(format, (p + 1))
         AND f = 0
 
         p := p + 1
         SWITCHON type INTO
         $( DEFAULT  :  wrch(type); ENDCASE
 
            CASE '$' :  GOTO r
            CASE 'S' :  f := writes;   GOTO l
            CASE 'C' :  f := wrch;     GOTO l
            CASE 'Z' :  f := wrid; t := t - 1;     GOTO l
            CASE 'I' :  f := writed;   GOTO m
            CASE 'N' :  f := writed;   GOTO l
         m: p := p + 1
            n := getbyte(format, p)
            n := '0' LE n LE '9' -> n - '0', n - 'A' + 10
         l: f(q, n)
         r: t := t + 1
         $)
      $)
      or wrch(k)
   $)
   wrch := save_wrch
   selectoutput(o)
$)

AND mdabort() BE
/*  Routine  called to abort the process after a fatal error detected by
the machine-dependent logic.  Prints  a  standard  error  message,  then
enters normal shutdown sequence.  */
$( mderpr("*NProcess aborted due to above error*N")
   mihalt(FALSE)
$)

AND mdfind() = VALOF
/*  Standard  ML/I  machine-dependent  function; the main hash function.
'idpt' points to the identifier  to  be  hashed  (an  unpacked  string).
'idlen'  is  its  length.  This version assumes 'lhv' is a power of two.
*/
$( MANIFEST $( lhvm1 = lhv - 1 $)

   LET r = idlen

   IF r > 2 THEN r := r + idpt!2
   r := r + idpt!0

   offset := ((r*113) >> 1) & lhvm1
   RESULTIS hashpt + offset
$)

AND mdread() = VALOF
/*  The  main  input  routine.   Handles input switching (using S10) and
input translation (using S16 and S17).  */
$( LET s10 = !at_s10
   AND rewflag_ = FALSE

   IF s10 = 0 RESULTIS endstreamch

   /*  Handle  legal values of S10 greater than 100. These indicate that
   100 should be subtracted from the value of S10, and the  result  used
   to  select  the  input  stream.   Additionally, the stream is rewound
   (repositioned at its start).  */

   IF 101 LE s10 LE 100 + maxinputs THEN
   $( s10 := s10 - 100
      !at_s10 := s10
      rewflag_ := TRUE
   $)

   /*  Check  for  values  of  S10 outside the permitted range.  Illegal
   values  of  S10  are  replaced  by  zero;  this  is  trapped  because
   input_streams!0 is zero.  */

   UNLESS 1 LE s10 LE maxinputs THEN s10 := 0

   $( LET strp = input_streams!s10

      IF strp = 0 THEN
      $( !at_s12 := !at_s12 + 10        // Ensure this message comes out
         prerr()
         mderpr("*NS10 has illegal value")
         mdconv(!at_s10)                // Convert S10 value to identifier
         prviz()
         mdabort()
      $)
      selectinput(strp)
   $)

   /* Rewind the stream if requested */

   IF rewflag_ THEN
   $( LET strp = rewind()

      IF strp < 0 THEN
      $( !at_s12 := !at_s12 + 10        // Ensure this message comes out
         prerr()
         mderpr("*NCannot rewind input stream %N*N", s10)
         mdabort()
      $)

      selectinput(strp)
   $)

   $( LET c = rdch()

      /*  Action  at  end  of  file.   If S10 is not the "revert stream"
      (whose value is in S23) then  input  is  switched  to  the  revert
      stream; otherwise the process is terminated.  */

      IF c = endstreamch THEN
      $( IF s10<maxinputs THEN !at_s23 := input_streams!(s10+1) \= 0 -> s10+1, s10
                                        // 'input_streams!0' is zero, as are unallocated streams
		 IF s10 = !at_s23 RESULTIS endstreamch
         !at_s10 := !at_s23
         RESULTIS mdread()
      $)

      IF c = !at_s16 THEN c := !at_s17
      UNLESS 0 LE c LE 127 THEN
      $( ersic()                        // Illegal input character
         c := error_character           // Substitute the 'error character'
      $)
      RESULTIS c
   $)
$)

AND mdconv(value) BE
/*  Converts  'value'  to  a  minimum-width,  unpacked character string,
pointed to by 'idpt'.  The string is stored in the area 'convarea' which
is allocated by the machine-independent logic.  The length of the string
is set into 'idlen'.  */
$( LET neg_ = FALSE

   idpt, idlen := convarea, 0

   IF value < 0 THEN
      value, neg_ := -value, TRUE

   TEST value = 0 THEN
   $( !idpt := '0'
      idlen := 1
      RETURN
   $)
   OR
   $( !idpt := value REM 10 + '0'
      idpt, idlen := idpt - 1, idlen + 1
      value := value/10
   $) REPEATUNTIL value = 0

   TEST neg_ THEN
      !idpt, idlen := '-', idlen + 1
   OR
      idpt := idpt + 1
$)
 
AND mdnum() = VALOF
/* Converts the unpacked string pointed at  by  'idpt'  into  a  decimal
number.   'idlen'  gives  the  length  of  the string.  Yields FALSE iff
'idpt' does not point at a valid digit.  */
$( LET c = !idpt

   meval := 0
   UNLESS '0' LE c LE '9' RESULTIS FALSE

   FOR i = idpt TO spt DO
   $( c := !i
      UNLESS '0' LE c LE '9' DO erlia()
      meval := meval*10 + c - '0'
   $)
   RESULTIS TRUE
$)

AND mdinit() BE
/* Routine to perform any machine-dependent initialisation which  cannot
be done until machine-independent initialisation has been completed.  */
$( nlpending_ := TRUE

// bmove := fmove                       // Share the code of this routine

   /*  Obtain  addresses  of commonly used system variables.  This is to
   improve efficiency.  */

   at_s10 := svarpt - 10
   at_s12 := svarpt - 12
   at_s16 := svarpt - 16
   at_s17 := svarpt - 17
   at_s19 := svarpt - 19
   at_s20 := svarpt - 20
   at_s21 := svarpt - 21
   at_s22 := svarpt - 22
   at_s23 := svarpt - 23

   /* Initialise system variables */

   !at_s10 := 1                         // Initial input from first input file
   !at_s12 := error_quota               // Initial lines limit on debugging file
   !at_s16 := -1                        // No input translation
   !at_s21 := 1                         // Initial output only to primary output
   !at_s23 := 1                         // Reverted input stream is initially first input stream
// svarpt!(-18) := b_uinfi(2) = 2 -> 1, 0
                                        // Print environment in batch only by default
   svarpt!(-18) := 1					// DC: b_uinfi(2) = 2 appears to indicate 'batch mode'
$)

AND mdfinal() BE
/* Routine to perform machine-dependent finalisation.  */
$( !at_s12 := maxint                    // Ensure statistics are output OK

   IF svarpt!(-18) NE 0 THEN prenv()
$)

AND fmove(len, src, dest) BE            // Code not shared by 'bmove'
/* Moves 'len' words from 'src' to 'dest', starting at the beginning (lowest address)
*/
// move(len*bytesperword, src*bytesperword, dest*bytesperword)
   FOR i=0 TO len-1 DO dest!i :=src!i

AND bmove(len, src, dest) BE            // Code not shared by 'fmove'
/* Moves 'len' words from 'src' to 'dest', starting at the end (highest address)
*/
// move(len*bytesperword, src*bytesperword, dest*bytesperword)
   FOR i=len-1 TO 0 by -1 DO dest!i :=src!i

AND mdouch(c) BE
/* Main output routine.  Handles multi-stream output, and output listing
if required.  */
$( LET s20 = !at_s20

   IF nlpending_ THEN
   $( !at_s19 := !at_s19 + 1            // Update output line number
      selectoutput(listing_output)
      IF s20 = 2 THEN writef("%I5.   ",!at_s19)
   $)

   nlpending_ := (c = '*N')

   IF s20 NE 0 THEN
   $( selectoutput(listing_output)
      $( LET ec = wrch(c)
         IF ec < 0 THEN write_error("listing", ec)
      $)
   $)

   IF !at_s21 NE 0 THEN
   $( selectoutput(primary_output)
      $( LET ec = wrch(c)
         IF ec < 0 THEN write_error("primary", ec)
      $)
   $)

   IF !at_s22 NE 0 THEN
   $( selectoutput(secondary_output)
      $( LET ec = wrch(c)
         IF ec < 0 THEN write_error("secondary", ec)
      $)
   $)
$)

AND write_error(stream_name, ec) BE
/* Routine called after a failure to write to one of the output streams.
A  suitable  message  is  sent to the debugging file, and the process is
aborted.  */
$( // ec := ABS ec
   !at_s12 := !at_s12 + 10              // Ensure this message comes out
   prerr()
   mderpr("*NError while writing to %S output file -%S", stream_name, ssmessage(ec, " "))
   mdabort()
$)

AND mddiv(a, b) = VALOF
/*  Divides  'a'  by  'b' according to the rules in Section 2.6.3 of the
User's Manual.   'b'  can  never  be  zero;  this  is  detected  by  the
machine-independent logic.  */
$( LET neg_ = FALSE
   AND res = 0

   IF a < 0 THEN
      a, neg_ := -a, NOT neg_

   IF b < 0 THEN
      b, neg_ := -b, NOT neg_

   res := a/b

   /* Round result towards minus infinity */

   IF neg_ & a REM b NE 0 THEN
      res := res + 1

   RESULTIS neg_ -> -res, res
$)

AND prompt(text) BE
$( LET op = output()

   IF input_streams!1 = sysin $(
      selectoutput(sysout)
      FOR i = 1 TO getbyte(text, 0) DO WRCH(getbyte(text, i))
      selectoutput(op)
	  $)
$)

AND ssmessage(code, text) = VALOF
$( LET t = VEC 20
   AND i, k, stl = 0, code, 0
   AND str = VEC 256/bytesperword
   AND diag = 0
   
   SWITCHON code INTO $(
      DEFAULT:	diag := "(undefined error code)";	ENDCASE
	  CASE 199:	diag := "File does not exist:";		ENDCASE
	  CASE 215:	diag := "Too many parameters";		ENDCASE
	  CASE 277:	diag := "Too many input files";		ENDCASE
	  CASE 999:	diag := "Can't assign workspace!";	ENDCASE
      $)
   
   IF code<(-MAXINT) DO $( t!i:=((code>>1) REM 10)*2; i:=i+1; code := code/10 $)	// Yep.  Ugly.
   IF code<0 DO k := -code
   t!i, k, i := k REM 10, k/10, i+1 REPEATUNTIL k=0
   FOR j=1 TO maxstrlength DO PUTBYTE(str, j, ' ')
   IF code<0 DO $(
	  stl := stl+1
	  PUTBYTE(str, stl, '-')
	  $)
   FOR j = i-1 TO 0 BY -1 DO $(				// plant the error code...
	  stl := stl+1
      PUTBYTE(str, stl, t!j+'0')
	  $)
   stl := stl+1								// (space)
   FOR j = 1 TO GETBYTE(diag, 0) DO $(
      stl := stl+1
	  IF stl > maxstrlength BREAK
	  PUTBYTE(str, stl, GETBYTE(diag, j))
      $)

   UNLESS text=0 DO $(						// check for any text string
      stl := stl+1							// ...(space)...
      FOR j = 1 TO GETBYTE(text, 0) DO $(	// ...then add any accompanying text
		 stl := stl+1
		 IF stl > maxstrlength BREAK
		 PUTBYTE(str, stl, GETBYTE(text, j))
		 $)
      $)
   PUTBYTE(str, 0, stl)
// writef("i=%N, str=%S, text=%S*n", i, str, text=0 -> "(no text)", text)
   
   RESULTIS str
$)

IF debugging_ THEN
/*  Routine called on a program error or abort, to dump the ML/I stacks.
*/
$( LET userpostmortem() BE
   $( LET n = 0
   
      LET dump(strt, fin) BE
      $( LET online = 8
         FOR i = strt TO fin DO
         $( IF online = 8 THEN
            $( online := 0
               writef("*N%X8:  ", i)
            $)
            writef("  %X8", !i)
            online := online + 1
         $)
      $)
   
      writef("*N*NBackwards stack:   (lfpt = %X8)*N", lfpt)
      n := lfpt - 30
      IF n < ffpt THEN n := ffpt
      dump(n, endpt)
      writef("*N*NForwards stack:    (ffpt = %X8)*N", ffpt)
      n := ffpt + 30
      IF n > lfpt THEN n := lfpt
      dump(stffpt, n)
      newline()
   $)
$)

 .

// End of file ML1_MDS

