
// Segment 2 - Scanning routine, and routines for dealing with each construction

GET "ML1HDR"

LET basic_scan() BE
/*  The  main  scanning  routine.   This routine is NOT recursive in the
normal BCPL sense;  however,  recursion  is  effected  by  stacking  the
scanning  description  block  (sdb)  and  returning to this routine at a
fixed closure point.  */
$( LET wnspt, wnidpt = 0, 0             // Temporary storage for 'spt' and 'idpt' while processing warning markers
   AND cinfpt = 0                       // Points at information block in primary construction
   AND extra = 0                        // Used in code for MC-PVAR operation macro
   AND opmac_type = 0                   // Switch used in processing operation macro calls

   scanlevel := level()                 // Set up closure point for resumption of scan

   $( masksw := hashpt!(lhv + 4) & glbwsw
   
      sbstpl()
      bespt := zeropt
   
   bsloop:
      gtatom()
      IF !idpt = endstreamch THEN
      $( ermtst()
         mihalt()
      $)
   
      IF nestlv NE 0 THEN               // Search delimiter chain
      $( htabpt := nullpt
         chanpt := delpt

         IF debugging_ THEN
            IF chanpt = nullpt DO macerr(1)

         $( chlink := !chanpt
            IF cmpare(chanpt) THEN tebest()
            chanpt := chanpt + chlink
         $) REPEATUNTIL chlink = endchn

         IF masksw = 5 GOTO enexcl      // Looking for exclusive delimiter
      $)
   
      /* Search chain of names */
   
   bsname:
      htabpt := mdfind()
      chanpt := !htabpt
   
      UNTIL chanpt = nullpt DO          // Find the best match
      $( IF cmpare(chanpt + 1) THEN tebest()
         chanpt := !chanpt
      $)
   
   bsswit:
      SWITCHON bestpl INTO
      $( CASE bp_stack:                 // Place atom on forwards stack
         sastak:
         $( LET lv4 = ffpt
            bumpff(idlen)
            fmove(idlen, idpt, lv4)
            arglen := arglen + idlen
            GOTO bsloop
         $)
   
         CASE bp_skip:                  // Action at skip name
            encall()
            masksw := !binfpt & 4       // Extract "matched" option
            IF skiplv = 0 THEN
            $( skiplv := nestlv
               copdsw := !binfpt & 1    // Extract "delimiter" option
               coptsw := !binfpt & 2    // Extract "text" option
            $)
   
            /* Skip delimiter */
   
         skdel:
            sbstpl()
            IF coptsw = 0 THEN bestpl := bp_null
                                        // Arrange to ignore text of skip
            idlen := bespt - idpt + 1
            UNLESS ressp() GOTO skd1
   
            /* End of skip */
   
            TEST decalv() & skiplv LE nestlv THEN
            $( /* End of nested skip */
               masksw := 4
            $)
            OR
            $( /* End of outermost skip */

               skiplv := 0
               IF copdsw = 0 LOOP
   
               /* Copy closing skip delimiter to output text */

            bstrex: l_bstrex:
               masksw := hashpt!(lhv + 4) & glbwsw
               sbstpl()
               bespt := zeropt
            $)
   
            IF idlen = 0 GOTO bsloop
            GOTO bsswit
   
            /* Non-closing skip delimiter */
   
         skd1:
            IF skiplv NE nestlv GOTO bsswit
            IF copdsw = 0 \/ skval NE 0 GOTO bsloop
            UNLESS oplev = 0 GOTO sastak
   
         CASE bp_copy:                  // Add atom to output text
            FOR i = 0 TO idlen - 1 DO
            $( LET c = idpt!i

               IF c = slrep LOOP
               mdouch(c)
            $)
   
         CASE bp_null:                  // Skip over text - do not add to output text
            GOTO bsloop
   
         CASE bp_warn:                  // Action at warning marker
            ressp()
            wnspt, wnidpt := spt, idpt
            idlen := 0
            IF gsatom() THEN
            $( bestpl := bp_awarn
               masksw := 1
               GOTO bsname
            $)
   
         CASE bp_awarn:
            IF svarpt!(-3) = 0 THEN ersnw()
                                        // Give error message, unless user has suppressed them
            spt, idpt := wnspt, wnidpt  // Restore pointers, and insert warning marker in output text
            idlen := spt - idpt + 1
            GOTO bstrex
   
         CASE bp_stop:                  // Action at stop marker
            spt := idpt - 1
            IF !idpt = '*N' THEN linect := linect - 1
            ermtst()
            LOOP
   
         CASE bp_macro:                 // Action at macro or insert name
         CASE bp_insert:
            masksw := hashpt!(lhv + 4) & glbwsw
            tempt := binfpt
            IF !tempt = strmk THEN      // Straight-scan macro
            $( masksw := 0
               tempt := tempt + 1
            $)
            IF nestlv = 0 THEN
            $( cinfpt := tempt + 1
               caltyp := !tempt
               IF caltyp = locmk & NOT ohsw THEN
                                        // Stack new hash table, if not already done at this level
               $( declf(1)
                  !lfpt := hashpt
                  stkhsh()
               $)
               declf(sdbsz)             // Make space to stack the SDB later on
               nargpt := lfpt
               argct := 0
               declf(1)
               !lfpt := idpt
               declf(1)
               !lfpt := bespt + 1
               skval := -skval - 1
            $)
            encall()
            bestpl := bp_null
            UNLESS ressp() GOTO bsloop
            ENDCASE
   
         CASE bp_sdelim:                // Action at secondary delimiter
         dello:
            IF skiplv NE 0 THEN
            $( delpt := binfpt + bindic
               GOTO skdel
            $)
            IF nestlv = 1 THEN
            $( argct := argct + 1
               declf(1)
               !lfpt := idpt
               declf(1)
               !lfpt := bespt + 1
            $)
            delpt := binfpt + bindic
            bestpl := bp_null
            UNLESS ressp() GOTO bsloop
            ENDCASE

         DEFAULT:
            macerr(2)
      $)
   
      /* Perform call */
   
      IF decalv() LOOP
      skval := -skval - 1
      declf(1)                          // Stack terminator to argument vector
      !lfpt := nullpt
      fmove(sdbsz, at_sdb, nargpt)

      IF caltyp < uinsmk THEN           // Test for insert or straight-scan macro
      $( IF skval NE 0 THEN             // Still searching for label - forget the call
         $( lfpt := nargpt + sdbsz
            LOOP
         $)
         invoct := invoct + 1

         /* Test for user macro or operation macro */

         IF caltyp < opmk THEN          // Call of user macro
         $( levl := levl + 1
            stoppt := caltyp + cinfpt - 1
                                        // Pointer to end of replacement text
            spt := !cinfpt + cinfpt - 1 // Pointer to start of replacement text
            dbugpt, dbugsw := mtchpt, db_repl
            argpt, stakpt := nargpt, nargpt
            prscan()

            /* Set up temporary variables */

            $( LET capacity = cinfpt!1  // Get capacity of macro from information block

               tvarpt := lfpt - 1
               declf(4)
               lfpt!0, lfpt!1, lfpt!2, lfpt!3 := levl - oplev, invoct, argct, capacity
                                        // Initialise T1, T2 and T3
               declf(capacity - 3)
            $)
            LOOP                        // Continue the scan (now of the replacement text)
         $)
      $)

      /* Perform operation macro call or insert */

      levl := levl + 1
      UNLESS oplev = 0 DO
      $( declf(opdbsz)
         fmove(opdbsz, at_opdb, lfpt)
      $)
      oplev := oplev + 1
      mhshpt, topspt, optyp := hashpt, lfpt, caltyp
      dbugsw := db_eval
      argno := 1
      inffpt := ffpt
      dbugpt, stakpt := nargpt, nargpt
   
      /* Test for insert */
   
      IF caltyp GE uinsmk THEN          // Perform insert
      $( sqnum := skval
         stkarg(sa_01); sa_01:
         UNLESS gargch() DO erlia()
         TEST !spt NE 'L' THEN
         $( LET written_switch_ = FALSE
            AND insw = db_eval

            IF sqnum NE 0 THEN          // Searching for label - ignore insert
            $( opexit()
               LOOP
            $)
            IF !spt = 'W' THEN          // 'Written' argument
            $( written_switch_ := TRUE
               UNLESS gargch() DO erlia()
            $)
         
            SWITCHON !spt INTO
            $( DEFAULT:                 // Case of macro expression
                  IF written_switch_ THEN erlia()
                  spt := spt - 1        // Backspace to first character of expression
                  getexp()
                  mdconv(meval)
                  opexit()
                  GOTO bstrex
         
               CASE 'D':
                  insw := db_delim
                  getexp()
                  IF meval = 0 & argpt NE nullpt ENDCASE
                  varpt := argpt
                  er1tst()              // Check that this delimiter actually exists
                  ENDCASE
         
               CASE 'B':
                  insw := db_subarg
         
               CASE 'A':
                  getexp()
                  varpt := argpt
                  er1tst()              // Check that this argument actually exists
            $)
         
            dbugpt := argpt
            argno := meval
            prarg(insw)
            IF written_switch_ THEN     // Written argument - do no further evaluation
            $( idpt := spt + 1
               idlen := stoppt - idpt
               opexit()
               GOTO bstrex
            $)
      
            IF optyp = pinsmk THEN hashpt := dbugpt!10
                                        // Protected insert - use value of 'hashpt' from previous level
            unopdb()
            lfpt := stakpt              // Partially collapse stack, while evaluating inserted text
            tvarpt := dbugpt!9          // Previous value of 'tvarpt'
            argpt := dbugpt!2           // Previous value of 'argpt'
         $)
         OR
         $( /* Case of label */
      
            getexp()
            IF meval LE 0 THEN erlme()  // Zero and negative labels may not be placed
            opexit()
            IF meval = skval THEN skval := 0
                                        // Label was being searched for - search is now over
         
            /* Remember label if necessary */
         
            IF levl = 0 LOOP            // Labels in source text are not remembered
            UNLESS sklab(meval) DO      // New label being defined - add to chain of label entries
            $( declf(4)
               lfpt!0, lfpt!1, lfpt!2, lfpt!3 := labpt, meval, spt - stoppt, linect
               labpt := lfpt
               LOOP
            $)
            IF !chanpt + stoppt NE spt THEN erlmd()
                                        // Already seen at different offset in current text - thus multiply-defined
         $)
         LOOP
      
         /* Action at end of piece of text */
      
      entext: l_entext:
         IF (nestlv = 0) \/ (db_oparg NE dbugsw NE db_subarg) GOTO entx1
      
         /* End of argument - look ahead for exclusive delimiter... */
      
         setpts(db_delim)
         stoppt := idpt + idlen
         spt := idpt - 1
         masksw := 5
         GOTO bsloop
      
         /* ...subsequently returning here */
   
      enexcl:
         stoppt := idpt
         IF bindic = exclmk GOTO dello
      
      entx1:
         ermtst()
         IF dbugsw = db_oparg THEN saexit()
         unsdb()
         LOOP
      $)
   
      /***************************************************
      *                                                  *
      *      Code for each individual operation macro    *
      *                                                  *
      ***************************************************/
   
      opmac_type := !cinfpt
      SWITCHON opmac_type INTO
      $(    /*************************
            * Code for M C A L T E R *
            *************************/
   
         CASE mc_alter:
            argno := 2
            stkarg(sa_02); sa_02:
            chatom()
            sqnum, sqpt := idlen, idpt
            argno := 1
            stkarg(sa_03); sa_03:
            chatom()
            UNLESS ludel(delchn) DO erlia()
                                        // Specified delimiter does not exist
            arglen := sqnum
            eriapt := sqpt
            argno := 2

            $( IF chlink - arglen LE 3 DO erlia()
                                        // New delimiter is too long
               spt := chanpt + 3        // Point to old delimiter
   
               /* Test for case of N (the node marker) or whatever has replaced it */
   
               IF spt = knpt THEN
               $( LET c = !sqpt         // New node marker

                  UNLESS 'A' LE c LE 'Z' \/
                         '0' LE c LE '9' DO erlia()
                                        // New node marker is not a letter or a digit
               $)
   
               chanpt!2 := arglen       // Set up length of new delimiter
               tempt := spt + arglen    // Point to where nextlink will have to go
   
               $( LET nlink = !infopt   // Old nextlink
   
                  UNLESS nlink = endchn DO
                                        // Relocate nextlink
                     nlink := nlink + infopt - tempt
                  !tempt := nlink
                  fmove(arglen, sqpt, spt)
                                        // Move in the new delimiter
                  chanpt := chanpt + chlink
                                        // Move past this instance of the delimiter
               $)
            $) REPEATWHILE ludel(chanpt)

            opexit()
            LOOP
   
            /*******************
            * Code for M C G O *
            *******************/
   
         CASE mc_go:
            UNLESS argct = 1 DO         // Conditional goto
            $( optyp := tesdel(dif)     // 0 for IF, 1 for UNLESS
               argno := 2
               meval := tesdel(dge)
               sqnum, sqsw := 0, FALSE
               SWITCHON meval INTO      // Type of condition
               $( CASE 4:               // Character comparison   (=)
                     stkarg(sa_04); sa_04:
                     sqnum := arglen
                     argno := 3
                     stkarg(sa_05); sa_05:
                     IF arglen NE sqnum ENDCASE
                                        // Conditional goto fails
                     tempt := inffpt
      
                     $( UNLESS advnce() DO
                                        // Conditional goto succeeds
                        $( optyp := optyp + 1
                           ENDCASE
                        $)
                        IF !spt NE !tempt ENDCASE
                                        // Conditional goto fails
                        tempt := tempt + 1
                     $) REPEAT
      
                  CASE 3:               // Belongs to class   (BC)
                     stkarg(sa_06); sa_06:
                     argno := 3
                     stkarg(sa_07); sa_07:
                     chekid()
                     UNLESS idlen = 1 DO erlia()
                                        // Class must be a single character
                     $( LET class = !idpt
                        idlen := eriapt - inffpt
                                        // Length of argument 2
                        IF idlen = 0 ENDCASE
                                        // Conditional goto fails
                        idpt := inffpt

                        SWITCHON class INTO
                        $( CASE 'N':    // Class of numbers
                           $( LET c = !idpt

                              WHILE (c = '+') \/ (c = '-') DO
                              $( idpt, idlen := idpt + 1, idlen - 1
                                 IF idlen = 0 ENDCASE
                                        // Conditional goto fails
                                 c := !idpt
                              $)
                              WHILE '0' LE c LE '9' DO
                              $( idpt, idlen := idpt + 1, idlen - 1
                                 IF idlen = 0 THEN
                                        // Conditional goto succeeds
                                 $( optyp := optyp + 1
                                    ENDCASE
                                 $)
                                 c := !idpt
                              $)
                              ENDCASE   // Conditional goto fails
                           $)

                           CASE 'L':    // Class of letters
                           CASE 'I':    // Class of identifiers
                              WHILE idlen > 0 DO
                              $( LET c = !idpt
                                 UNLESS 'A' LE c LE 'Z' \/
                                        'a' LE c LE 'z' \/
                                        (('0' LE c LE '9') & (class = 'I')) ENDCASE
                                        // Conditional goto fails
                                 idpt, idlen := idpt + 1, idlen - 1
                              $)
                              optyp := optyp + 1
                                        // Conditional goto succeeds
                              ENDCASE

                           DEFAULT:
                              erlia()   // Illegal class
                        $)
                     $)
                     ENDCASE
      
                  CASE 0:               // Numerical comparison   (GE)
                     sqnum := 1
      
                  CASE 1:               // Numerical comparison   (GR)
                     sqsw := TRUE
      
                  CASE 2:               // Numerical comparison   (EN)
                     stkarg(sa_08); sa_08:
                     getexp()
                     sqnum := sqnum + meval
                     argno := 3
                     stkarg(sa_09); sa_09:
                     getexp()
                     TEST sqsw THEN
                     $( IF meval < sqnum THEN
                           optyp := optyp + 1
                                        // Conditional goto succeeds
                     $)
                     OR
                     $( IF meval = sqnum DO
                                        // Conditional goto succeeds
                           optyp := optyp + 1
                     $)
                     ENDCASE

                  DEFAULT:
                     macerr(3)
               $)
   
               IF optyp NE 1 THEN
               $( opexit()
                  LOOP
               $)
               argno := 1
            $)
   
            /* Perform goto */
   
            stkarg(sa_10); sa_10:
            UNLESS gargch() & (!spt = 'L') DO erlia()
            getexp()
            IF meval = 0 THEN           // Label L0 - return from current level
            $( IF levl = 1 THEN erlme()
               opexit()
               IF dbugsw = db_oparg THEN saexit()
               unsdb()
               LOOP
            $)
            IF meval < 0 THEN erlme()
            opexit()
            TEST sklab(meval) THEN      // Label already encountered
               spt, linect := !chanpt + stoppt, chanpt!1
            OR
               skval, sklin := meval, mchlin
            LOOP
   
            /***********************
            * Code for M C L E N G *
            ***********************/
   
         CASE mc_leng:
            stkarg(sa_11); sa_11:
            mdconv(arglen)              // Convert argument length to a string
            opexit()
            GOTO bstrex
   
            /***********************
            * Code for M C N O T E *
            ***********************/
   
         CASE mc_note:
            stkarg(sa_12); sa_12:
            idpt, idlen := eriapt, arglen
            mderpr("*N%Z")
            opexit()
            TEST svarpt!(-4) = 0 THEN prctxt() OR mderpr("*N")
            LOOP
   
            /***********************
            * Code for M C P V A R *
            ***********************/
   
         CASE mc_pvar:
            stkarg(sa_13); sa_13:
            getexp()
            extra := meval - pvnum
            IF extra > 0 THEN
            $( mkroom(extra)
               FOR i = 0 TO extra - 1 DO ndefpt!i := 0
                                        // Zeroise new permanent variables
               pvnum := meval
               !pvarpt := pvnum
            $)
            opexit()
            LOOP
   
            /*********************
            * Code for M C S E T *
            *********************/
   
         CASE mc_set:
            stkarg(sa_14); sa_14:
            chekid()
            spt := idpt + idlen - 1
            sqpt := gmeadd()            // Get address of macro variable on LHS
            argno := 2
            stkarg(sa_15); sa_15:
            getexp()
            !sqpt := meval              // Update the variable
            opexit()
            LOOP
   
            /*********************
            * Code for M C S U B *
            *********************/
   
         CASE mc_sub:
            stkarg(sa_16); sa_16:
            optyp, argno := arglen, 2
            stkarg(sa_17); sa_17:
            getexp()
            sqnum := 0
            subchk()
            sqnum := meval - 1
            argno := 3
            stkarg(sa_18); sa_18:
            getexp()
            subchk()
            idlen := meval - sqnum
            idpt := inffpt + sqnum
            opexit()
            GOTO bstrex
   
            /**********************************
            * Code for M C N O - - - - macros *
            **********************************/
   
         CASE mc_nowarn:
            hashpt!(lhv + 4) := 7
   
         CASE mc_nodef:
         CASE mc_noins:
         CASE mc_noskip:
            hashpt!(lhv + opmac_type) := dbugpt + sdbsz
                                        // Set new limit on appropriate local constructions
            opexit()
            LOOP
   
            /***********************
            * Code for M C W A R N *
            ***********************/
   
         CASE mc_warn:
            ntypsw := ty_warn
            ENDCASE
   
            /***********************
            * Code for M C S T O P *
            ***********************/
   
         CASE mc_stop:
            ntypsw := ty_stop
            ENDCASE
   
            /*********************
            * Code for M C D E F *
            *********************/
   
         CASE mc_def:
            argno := argct - 1
            sqsw := (tesdel(das) NE 0)  // Indicates whether macro is to be straight-scan
            sqnum := 3                  // Default capacity of macro
            IF argct NE 2 THEN          // The VARS part is present
            $( argno := 1
               stkarg(sa_19); sa_19:
               getexp()
               IF meval > 3 THEN sqnum := meval
                                        // User-specified capacity for macro
               ffpt := inffpt
            $)
            argno := !dbugpt
            stkarg(sa_20); sa_20:
            argno := argno - 1
            ntypsw := ty_macro
            ENDCASE
   
            /*********************
            * Code for M C I N S *
            *********************/
   
         CASE mc_ins:
            sqnum := pinsmk
            ntypsw := ty_insert
            IF argct NE 1 THEN          // Insert type has been specified
            $( stkarg(sa_21); sa_21:
               chekid()
               UNLESS idlen = 1 DO erlia()
               $( LET c = !idpt
      
                  IF c = 'P' THEN
                  $( argno := 2
                     ENDCASE
                  $)
                  UNLESS c = 'U' DO erlia()
               $)
               sqnum := uinsmk
               argno := 2
            $)
            ENDCASE
   
            /***********************
            * Code for M C S K I P *
            ***********************/
   
         CASE mc_skip:
            sqsw := 0
            ntypsw := ty_skip
            IF argct = 1 ENDCASE
            stkarg(sa_22); sa_22:
            WHILE gargch() DO
            $( SWITCHON !spt INTO
               $( DEFAULT:   erlia()
   
                  CASE 'D':  sqsw := sqsw \/ 1; LOOP
   
                  CASE 'T':  sqsw := sqsw \/ 2; LOOP
   
                  CASE 'M':  sqsw := sqsw \/ 4; LOOP
               $)
            $)
            argno := 2
            ENDCASE

         DEFAULT:
            macerr(4)
      $)
   
      /* Deal with structure representation */
   
      stkarg(sa_23); sa_23:
      nodept, optlev, delct := nullpt, 0, 0
      bumpff(1)
      exitsw, keysw, nodesw, consw := FALSE, TRUE, TRUE, TRUE
      lnodpt := lfpt
      getdel()
   
      /*  Start  building  up  the  information  block   for   the   new
      construction, on the forwards stack.  */

   ev1:
      !ffpt := ntypsw
      bumpff(1)
      SWITCHON ntypsw INTO
      $( CASE ty_macro:
            IF sqsw THEN
            $( !ffpt := strmk
               bumpff(1)
            $)
   
            /* Stack offsets of end and start of replacement text */
   
            !ffpt := stoppt - ffpt
            bumpff(1)
            !ffpt := inffpt + arglen - ffpt
            bumpff(1)
   
         CASE ty_insert:
            !ffpt := sqnum
            bumpff(1)
            ENDCASE
   
         CASE ty_skip:
            !ffpt := sqsw
            bumpff(1)
   
         CASE ty_stop:
         CASE ty_warn:
            ENDCASE

         DEFAULT:
            macerr(5)
      $)

      getdel()  REPEAT                  // This loop terminates at the end of the structure representation - see below
   
      /* Action at OPT */
   
   evopt: l_evopt:
      optlev := optlev + 1
      declf(allsz)                      // Stack the ALL block
      fmove(allsz, at_all, lfpt)
      optpt := ffpt
      opthpt := optpt
      allpt := nullpt
      nodesw := FALSE
      getdel()
   
   evcont:
      IF delct = 0 GOTO ev1
      getdel()  REPEAT                  // Jumps to 'evexit' at end of structure representation

      /* Action at OR */
   
   evor: l_evor:
      joinch()
      IF delct = 0 THEN bumpff(1)
      nodept := optpt
      optpt := ffpt
      nodesw := TRUE
      getdel()
   
      /* Check that name of branch is unique */
   
      chanpt := opthpt

      IF debugging_ THEN
         IF chanpt = nullpt DO macerr(6)

      $( chlink := !chanpt
         IF chanpt = optpt GOTO evcont
         idpt := optpt
         infopt := chanpt
   
         $( idlen := idpt!1
            idpt := idpt + 2
            UNLESS cmpare(infopt) BREAK
            idpt := idpt + idlen
      
            /* Check WITH, WITHS and SPACES */
      
            tespac()
            tempt := infopt
            infopt := idpt
            idpt := tempt
            tespac()
            UNLESS tewith(infopt) DO
            $( IF tewith(idpt) BREAK
               erlia()
            $)
         $) REPEATWHILE tewith(idpt)
   
         IF chlink = endchn BREAK
         chanpt := chanpt + chlink
      $) REPEAT
   
      /* Action at end of structure representation */
   
   evexit: l_evexit:
      IF delct = 0 \/ optlev NE 0 THEN erlia()
      chanpt := nodept
      TEST chanpt NE nullpt THEN
      $( chlink := !chanpt
         !chanpt := endchn
         IF chlink = endchn BREAK
         chanpt := chanpt + chlink
      $) REPEAT
      OR GOTO evex1
   
      /* Check that all nodes have been placed */
   
      UNTIL lfpt = topspt DO
      $( IF lfpt!1 = 0 THEN erlia()
         lfpt := lfpt + nodesz
         LOOP
      
      evex1:
         UNLESS exitsw DO erlia()
      $)
   
      /* Action after processing of structure representation */
   
      offset := 1
      idlen := ffpt - stoppt
   
      SWITCHON ntypsw INTO
      $( CASE ty_macro:                 // Macro case - make delimiter structure adjacent to replacement text
            fmove(idlen, stoppt, eriapt)
            stoppt := inffpt
            ffpt := ffpt - arglen
            offset := offset + eriapt - stoppt
            idlen := ffpt - stoppt
            ENDCASE

         CASE ty_warn:
            UNLESS delct = 1 DO erlia()
            TEST optyp NE opmk THEN hashpt!(lhv + 4) := 6 OR glbwsw := 6
            ENDCASE

         CASE ty_stop:
            UNLESS delct = 1 DO erlia()
            ENDCASE

         CASE ty_insert:
            UNLESS delct = 2 DO erlia()
   
         CASE ty_skip:
            ENDCASE
   
         DEFAULT:
            macerr(7)
      $)
   
      /* Add to environment */
   
      idpt := stoppt
      TEST optyp = locmk THEN           // Local case - make space on backwards stack
      $( opexit()
         declf(idlen)
         ndefpt := lfpt
      $)
      OR
      $( mkroom(idlen)                  // Global case - make space on forwards stack
         idpt := idpt + idlen           // Relocate pointer after text has been shuffled up
         opexit()
      $)
      bespt := spt
      bmove(idlen, idpt, ndefpt)        // Move new definition into place
   
      chanpt := ndefpt + offset

      IF debugging_ THEN
         IF chanpt = nullpt DO macerr(8)

      $( chlink := !chanpt
         idpt := chanpt + 1
         idlen := !idpt
         idpt := idpt + 1
         spt := idpt + idlen - 1
         htabpt := mdfind()
         IF ndefpt NE lfpt THEN         // Global case - find end of locals on hash chain
         $( IF lfpt > !htabpt BREAK
            IF !htabpt > endpt BREAK
            htabpt := !htabpt
         $) REPEAT
   
         tempt := chanpt - 1
         !tempt := !htabpt
         !htabpt := tempt
         IF ndefpt NE lfpt THEN
         $( LET tpt = hashpt
   
            /* Add globals to all hash tables */
   
            $( tpt := tpt!(lhv + 5)
               IF tpt = nullpt BREAK
               htabpt := tpt + offset
               IF !htabpt = !tempt THEN
                  !htabpt := tempt
            $) REPEAT
         $)
         IF chlink = endchn BREAK
         chanpt := chanpt + chlink
      $) REPEAT
   
      spt := bespt
bsnext: l_bsnext:
   $) REPEAT                            // End of main scanning loop
$)

AND stkarg(label) BE
/*  The  entry  part of the linkroutine.  Evaluates argument 'argno' and
stacks the evaluated form; This is done by re-entering the main scanning
loop.  The parameter 'label' specifies the link to the call of 'stkarg';
it  is  not   strictly   necessary,   but   means   that   'stkarg'   is
machine-independent  (it  need have no knowledge of how return links are
stored on a particular implementation of BCPL).  */
$( linkpt := label                      // Save link in SDB

   prarg(db_roparg)
   arglen := 0

   longjump(scanlevel, bsnext)          // Resume scanning
$)

AND saexit() BE
/*  The exit part of the linkroutine.  This is called when the evaluated
form of the required argument has been stacked.   Resets  pointers,  and
resumes  the  scan at the point where the corresponding call of 'stkarg'
was made.  It is, however, important to realise that return is  made  to
the  closure 'scanlevel'; thus the code calling 'stkarg' should NOT have
any local variables extant in addition to those declared at the start of
'basic_scan', as they will be corrupted during argument evaluation.  */
$( lfpt, stoppt := topspt, ffpt
   eriapt := stoppt - arglen
   spt := eriapt - 1
   dbugsw := db_eval

   longjump(scanlevel, linkpt)          // Resume processing of outer call
$)

 .
