
// Segment 3 - Main subroutines (in alphabetical order)

GET "ML1HDR"

LET advnce() = VALOF
/*  Function  to  advance  to next character in the current text; yields
FALSE iff at end of current text.  */
$( spt := spt + 1
   IF spt = stoppt RESULTIS FALSE
   UNLESS spt = ffpt RESULTIS TRUE

   /* Stack is empty, and source text is being scanned - read in more text */

   $( LET c = mdread()

      IF nlsw_ & c NE endstreamch THEN  // Avoid lone endstreamch being treated as a line
      $( let n = !at_s2 + 1
   
         !at_s2 := n                    // Update S2
   
         IF tlinct = linect THEN tlinct := n
         linect := n
   
         IF !at_s1 NE 0 THEN            // Insert startline
         $( !ffpt := slrep
            bumpff(1)
         $)
      $)

      TEST c = endstreamch THEN         // End of input
      $( ffpt := spt
         UNLESS nlsw_ DO                // Add preceding newline, if necessary
         $( !ffpt := '*N'
            bumpff(1)
         $)
         !ffpt := endstreamch
         bumpff(1)
      $)
      OR
      $( nlsw_ := (c = '*N')
         !ffpt := c
         bumpff(1)
      $)
   $)
   RESULTIS TRUE
$)

AND bumpff(n) BE
/* Routine to increase ffpt, checking for stack overflow.  */
$( ffpt := ffpt + n
   IF ffpt GE lfpt THEN erlso()
$)

AND chatom() BE
/* Routine to check that the argument described by argno and  arglen  is
exactly one atom.  */
$( IF arglen = 0 THEN erlia()

   gtatom()

   IF idlen NE arglen THEN erlia()
$)

AND chekid() BE
/*  Routine  to  check  that the current text consists (after removal of
leading and trailing spaces) of exactly one atom.  */
$( UNLESS gsatom() DO erlia()
   IF gsatom() THEN erlia()
$)

AND ckvaly(ptr, type) = VALOF
/* Function to check  whether  a  specified  construction  is  currently
valid; yields TRUE iff it is.  */
$( IF ptr > endpt RESULTIS TRUE         // Outside stacks - must be an operation macro
   RESULTIS ptr < hashpt!(lhv + type - 1)
$)

AND cmpare(ptr) = VALOF
/* Function to compare two atoms; yields FALSE iff comparison fails.  On
return, 'infopt' points beyond the matched atom.  */
$( ptr := ptr + 1
   IF idlen NE !ptr RESULTIS FALSE
   ptr := ptr + 1
   infopt := ptr + idlen
   FOR i = 0 TO idlen - 1 DO
      UNLESS ptr!i = idpt!i RESULTIS FALSE
   RESULTIS TRUE
$)

AND corect(n, ptr) BE
/* Routine to relocate a pointer - called from 'mkroom'.  */
   IF ndefpt LE !ptr LE inffpt THEN !ptr := !ptr + n

AND decalv() = VALOF
/* Routine to restore previously stacked scan information at the end  of
a nested construction.  Yields FALSE iff level of calls has become zero.
*/
$( nestlv := nestlv - 1
   IF nestlv = 0 RESULTIS FALSE
   delpt, mtchpt, mchlin := lfpt!0, lfpt!1, lfpt!2
   lfpt := lfpt + 3
   cllfpt := lfpt
   RESULTIS TRUE
$)

AND declf(n) BE
/* Routine to decrease 'lfpt', checking for stack overflow.  */
$( lfpt := lfpt - n
   IF ffpt GE lfpt THEN erlso()
$)

AND er1tst() BE
/*  Routine  to  check  the validity of a macro element, as described by
'varpt' and 'meval'; gives error if not valid.  */
$( IF (varpt = nullpt) \/ (meval > !varpt) \/ (meval LE 0) THEN
      erlme()
$)

AND encall() BE
/* Routine to stack and re-initialise scanning information when a nested
construction is encountered.  'cllfpt' is set to  point  to  the  latest
block  of  information stacked (this is used by 'prmiss' to "back up" if
stack overflow occurs).  */
$( IF nestlv NE 0 THEN
   $( declf(3)
      lfpt!0, lfpt!1, lfpt!2 := delpt, mtchpt, mchlin
   $)
   cllfpt := lfpt
   delpt := binfpt + bindic - 2
   mtchpt, mchlin, nestlv := bfndpt, linect, nestlv + 1
$)

AND gargch() = VALOF
/* Function to set 'spt' to point at first non-space  character  in  the
current text; yields FALSE iff end of current text reached.  */
$( WHILE advnce() DO
      IF !spt NE '*S' RESULTIS TRUE
   RESULTIS FALSE
$)

AND getexp() BE
/* Routine to read an expression from the current text, and evaluate it.
Constants and variables are permitted; evaluation is from left to  right
with no operator precedence.  */
$( LET varsw, sum, olidpt = TRUE, 0, spt
   AND opsw, negval_, op1 = 0, ?, ?

   meval := 0
   negval_ := FALSE

   WHILE gsatom() DO
   $( LET op = !idpt
      SWITCHON op INTO
      $( DEFAULT:
            UNLESS varsw DO erlia()
            UNLESS mdnum() DO gmeadd()
            IF negval_ THEN meval := -meval
            varsw := FALSE
            SWITCHON opsw INTO
            $( CASE 1:
                  meval := op1*meval

               CASE 0:
                  ENDCASE

               CASE 2:
                  IF meval = 0 THEN erlovf()
                  meval := mddiv(op1, meval)
                  ENDCASE

               DEFAULT:
                  macerr(9)
            $)
            opsw := 0
            negval_ := FALSE
            LOOP

            /* Minus and plus */

         CASE '-':
            negval_ := NOT negval_

         CASE '+':
            sum := sum + meval
            varsw := TRUE
            meval := 0
            LOOP

            /* Divide and times */

         CASE '/':
         CASE '**':
            opsw := op = '**' -> 1, 2
            IF varsw THEN erlia()
            op1 := meval
            varsw, meval := TRUE, 0
            LOOP
      $)
   $)
   IF varsw THEN erlia()
   meval := meval + sum
   idpt := olidpt
$)

AND gmeadd() = VALOF
/*  Function  to  decode  a  (possibly subscripted) macro variable name;
yields the address of the macro variable. 'meval'  is  set  to  previous
contents.  */
$( LET flagpt = idpt
   AND res = ?

   $( LET c = !idpt
      TEST (c NE 'P') & (c NE 'S') & (c NE 'T') THEN
      $( IF idpt = flagpt BREAK
         UNLESS mdnum() BREAK
         $( idpt := idpt - 1
            c := !idpt
            varpt := (c = 'T' -> tvarpt, c = 'S' -> svarpt, pvarpt)
            er1tst()
            res := varpt - meval
            meval := varpt!(-meval)
            IF idpt = flagpt RESULTIS res
         $) REPEAT
      $)
      OR
      $( IF idpt = spt BREAK
         idpt := idpt + 1
      $)
   $) REPEAT
   erlia()
$)

AND gsatom() = VALOF
/*  Function  to  extract the next non-space atom from the current text.
On exit, 'idpt' and 'idlen' describe the atom.  Yields FALSE iff at  end
of current text.  */
$( IF gargch() THEN
   $( UNLESS !spt = endstreamch DO
      $( spt := spt - 1
         gtatom()
         RESULTIS TRUE
      $)
   $)
   spt := spt - 1
   RESULTIS FALSE
$)

AND gtatom() BE
/*  Routine  to  extract  the  next atom from the current text. On exit,
'idpt' and 'idlen' describe the atom.  */
$( IF (levl = 0) & (skval GE 0) & (ffpt - spt = 1) THEN
                                        // Clear redundant information from forwards stack
   $( ffpt := inffpt
      spt := ffpt - 1
   $)

   UNLESS advnce() DO longjump(scanlevel, entext)

   IF !spt = '*N' THEN linect := linect + 1

   idpt, idlen := spt, 1
   $( LET c = !spt

      IF 'A' LE c LE 'Z' \/ 'a' LE c LE 'z' \/ '0' LE c LE '9' THEN
      $( WHILE advnce() DO
         $( c := !spt
            UNLESS 'A' LE c LE 'Z' \/ 'a' LE c LE 'z' \/ '0' LE c LE '9' BREAK
            idlen := idlen + 1
         $)
   
         /* End of identifier */
   
         spt := spt - 1                 // Backspace over character which terminated the atom
      $)
   $)
$)

AND ludel(ptr) = VALOF
/*  Function  to search the chain headed by 'ptr' for the atom described
by 'idpt' and 'idlen'; yields TRUE iff it is found.  */
$( chanpt := ptr

   IF debugging_ THEN
      IF chanpt = nullpt DO macerr(10)

   $( chlink := !chanpt
      IF cmpare(chanpt + 1) RESULTIS TRUE
      IF chlink = endchn RESULTIS FALSE
      chanpt := chanpt + chlink
   $) REPEAT
$)

AND lulayk(pridcall_) = VALOF
/*  Function  to  yield  the  address  of the entry for the keyword of a
layout character, or zero if not found.  */
$( LET ptr = laychn
   AND tptr, cptr = ?, ?

   $( tptr := ptr + 1                   // Point to LID
      cptr := !tptr + tptr              // Point to actual character

      /* Compare with list of layout characters, unless called from 'prid' routine */

      IF NOT pridcall_ & cmpare(tptr) THEN
      $( idpt, idlen := cptr, 1
         RESULTIS 0
      $)

      IF !cptr = !idpt RESULTIS ptr
      IF ptr = kspacs RESULTIS 0        // All the layout characters have been tried

      ptr := ptr + !ptr
   $) REPEAT
$)

AND mkroom(n) BE
/*  Routine  to create space on the forwards stack immediately above the
permanent variables (below the global definitions).  The number of extra
cells required is given by 'n'.  Used to make space  for  a  new  global
definition, or for additional permanent variables.  */
$( LET tpt, tsw = ?, ?

   ndefpt := pvarpt - pvnum             // Pointer to space about to be created
   bumpff(n)                            // Make space at top of forwards stack
   bmove(ffpt - ndefpt - n, ndefpt, ndefpt + n)
   dbugpt := stakpt

   $( tsw := dbugpt!8                   // Previous dbugsw
      IF tsw = db_repl BREAK
      corect(n, dbugpt + 4)             // Relocate 'spt'
      corect(n, dbugpt + 14)            // Relocate inffpt
      IF tsw = db_source GOTO mrexit
      corect(n, dbugpt + 12)            // Relocate stoppt
      IF tsw = db_oparg THEN
      $( tpt := dbugpt!3
         GOTO mkr2
      $)
   mkr1:
      dbugpt := dbugpt!1                // Previous stakpt
   $) REPEAT

   tpt := dbugpt!2                      // Previous 'argpt'
mkr2:
   tpt := tpt - 1
   IF !tpt = nullpt GOTO mkr1
   corect(n, tpt)
   GOTO mkr2
mrexit:
   pvarpt := pvarpt + n
   inffpt := inffpt + n
$)

AND opexit() BE
$( unopdb()
   unsdb()
$)

AND prarg(sw) BE
$( hashpt := mhshpt
   setpts(sw)
   spt := idpt - 1
   stoppt := idpt + idlen
   TEST sw = db_eval \/ sw = db_roparg THEN
                                        // Delete spaces if A or operation macro argument
   $( dbugsw := sw = db_eval -> db_subarg, db_oparg
      IF gsatom() THEN
      $( spt := idpt - 1
         stoppt := stoppt - 1  REPEATWHILE !stoppt = '*S'
         stoppt := stoppt + 1
      $)
   $) OR dbugsw := sw
   prscan()
$)

AND prscan() BE
/*  Routine  to initialise variables in SDB, before scanning a new piece
of text.  */
$( ohsw := FALSE
   skval := 0
   linect := 1
   labpt := nullpt
$)

AND ressp() = VALOF
/* Yields FALSE iff not at end of call.  */
$( spt := bespt
   bespt := zeropt
   IF bindic NE exclmk THEN
   $( linect := beslin
      RESULTIS bindic = endchn
   $)

   /* Exclusive delimiter */

   idlen := 0
   bindic := endchn
   spt := idpt - 1
   IF !idpt = '*N' THEN linect := linect - 1
   RESULTIS TRUE
$)

AND sbstpl() BE
/*  Routine  to  set  'bestpl',  the  switch which determines what is to
happen to scanned text.  */
$( bestpl := bp_copy                    // By default, copy text to output
   IF oplev NE 0 THEN bestpl := bp_stack
                                        // If scanning for delimiter, stack text instead
   IF skval NE 0 THEN bestpl := bp_null
                                        // If scanning for label, discard text
$)

AND setpts(sw) BE
/*  Routine to set up idpt and idlen to describe the argument (if 'sw' =
'db_roparg' or  'db_subarg'  or  'db_eval')  or  delimiter  (if  'sw'  =
'db_delim') specified by 'argno'.  */
$( LET n = argno*2
   IF sw = db_delim THEN n := n + 1
   IF debugging_ THEN
      UNLESS sw = db_roparg \/ sw = db_subarg \/ sw = db_eval DO macerr(11)
   idpt := !(dbugpt - n)
   idlen := !(dbugpt - n - 1) - idpt
$)

AND sklab(l) = VALOF
/* Function to determine whether a given label is present in the current
text; yields TRUE iff it is. 'chanpt' is set to the offset of the  label
from the end of the text.  */
$( chanpt := labpt
   UNTIL chanpt = nullpt DO
   $( IF chanpt!1 = l THEN
      $( chanpt := chanpt + 2
         RESULTIS TRUE
      $)
      chanpt := !chanpt
   $)
   RESULTIS FALSE
$)

AND stkhsh() BE
/* Routine to stack the current hash table on the backwards stack.  */
$( ohsw := TRUE
   declf(lhv + 5)
   fmove(lhv + 5, hashpt, lfpt)
   hashpt := lfpt
$)

AND subchk() BE
$( IF meval LE optyp THEN
   $( IF meval LE 0 THEN
         meval := meval + optyp
      IF meval > sqnum RETURN
   $)

   /* MC-SUB has null value - perform null insert */

   idlen := 0
   opexit()
   longjump(scanlevel, bstrex)
$)

AND tebest() BE
$( LET tidpt, tspt = idpt, spt          // Save current position while looking ahead
   AND fndpt = infopt - idlen - 2       // Point to orlink
   AND type = ?

   tlinct := linect

   $( IF !infopt = spcsmk THEN          // This possibility ends with SPACES - move over any spaces in current text
      $( infopt := infopt + 1
         gargch()
         spt := spt - 1                 // Move back over last (non-space) character
      $)

      IF tewith(infopt) THEN            // Possibility contains WITH or WITHS
      $(
      tbwith:
         UNLESS advnce() BREAK          // End of the current text - no match
         spt := spt - 1
         gtatom()
         IF (!idpt = '*S') & (indic = wthsmk) GOTO tbwith
         UNLESS cmpare(infopt) BREAK    // Subsequent atom does not match
         LOOP
      $)

      TEST indic = exclmk THEN          // Exclusive delimiter
         IF bindic NE exclmk GOTO tb3
      OR
         IF bindic = exclmk BREAK       // Not exclusive delimiter
      IF bespt GE spt BREAK             // A longer construction name has already been found
   
   tb3:
      TEST htabpt = nullpt THEN
      $( /* Secondary delimiter case */
   
         bestpl := bp_sdelim
      $)
      OR
      $( infopt := infopt + 1
         TEST !infopt = ty_stop THEN    // Case of stop marker
         $( IF (levl NE 0) \/ (skval = 0 = skiplv) BREAK
            bestpl := bp_stop
         $)
         OR
         $( IF (!infopt & masksw) = 0 BREAK
            type := !infopt
            IF (type = ty_insert) & (masksw = 1) BREAK
            UNLESS ckvaly(chanpt, type) BREAK
                                        // This name is not currently valid
            infopt := infopt + 1
            bestpl := type
         $)
      $)

      bfndpt := fndpt
      binfpt, bindic, bespt, beslin := infopt, indic, spt, linect
      BREAK
   $) REPEAT

   idpt, spt := tidpt, tspt
   idlen := spt - idpt + 1
   linect := tlinct
$)

AND tesdel(ptr) = VALOF
/* Function to search a specified secondary delimiter chain, looking for
a match; yields the position within chain of the matched delimiter.  */
$( LET item = 0

   setpts(db_delim)
   chanpt := ptr

   IF debugging_ THEN
      IF chanpt = nullpt DO macerr(12)

   $( chlink := !chanpt
      IF cmpare(chanpt) RESULTIS item
      item := item + 1
      IF chlink = endchn THEN macerr(13)
      chanpt := chanpt + chlink
   $) REPEAT
$)

AND tewith(ptr) = VALOF
/* Function to check if the item pointed at by its parameter is 'withmk'
or  'wthsmk'; yields TRUE iff so; indic is set to the item.  Yields TRUE
iff WITH or WITHS found.  */
$( indic := !ptr
   RESULTIS (indic = withmk) \/ (indic = wthsmk)
$)

AND unopdb() BE
$( ffpt := inffpt
   levl := levl - 1
   oplev := oplev - 1
   IF oplev NE 0 THEN
   $( LET tpt = topspt

      fmove(opdbsz, tpt, at_opdb)
      lfpt := tpt + opdbsz
   $)
$)

AND unsdb() BE
$( IF dbugsw = db_repl THEN levl := levl - 1
   tempt := stakpt
   fmove(sdbsz, tempt, at_sdb)
   lfpt := tempt + sdbsz

   /* Delete text of call if in source text */

   IF levl = 0 THEN
   $( LET size = ffpt - spt - 1

      spt := spt + 1
      IF size NE 0 THEN fmove(size, spt, inffpt)
      spt := inffpt - 1
      ffpt := inffpt + size
   $)
$)

 .
