
// Segment 5 - Error routines (alphabetical order)

GET "ML1HDR"

LET erlmd() BE
/* Multiply-defined label.  */
$( prerr()
   mderpr("Label %N is multiply-defined", meval)
   prctxt()
$)

AND erlme() BE
/* Illegal macro element.  */
$( prerr()
   idlen := 1
   mderpr("%Z%N is illegal macro element", meval)
   prctxt()
   mcabrt()
$)

AND erlia() BE
/* Illegal argument.  */
$( prerr()
   mderpr("Argument %N has illegal value", argno)
   idlen := arglen
   idpt := eriapt
   prviz()
   mcabrt()
$)

AND erlovf() BE
/* Arithmetic overflow.  */
$( prerr()
   mderpr("Arithmetic overflow")
   prctxt()
   mcabrt()
$)

AND erlso() BE
/* Stack overflow.  If the current text is the  source  text,  then  the
following   additional   information   is   given:   if  there  are  any
constructions currently unmatched, or if a search is being  made  for  a
label  as  a  result  of  a  forward  MC-GO, then appropriate diagnostic
messages are printed.  */
$( prerr()
   mderpr("Process aborted for lack of storage")
   TEST (dbugsw = db_source) & ((skiplv NE 0) \/ (skval NE 0)) THEN
   $( mderpr(" possibly due to*N")
      prmiss()
   $)
   OR prctxt()
   mihalt()
$)

AND ermtst() BE
/* Routine to test for mismatches. Prints appropriate diagnostics if any
are found. */
$( UNLESS skiplv = 0 = skval DO
   $( prerr()
      prmiss()
   $)
$)

AND ersic() BE
/* Illegal input character.  */
$( LET oidpt, oidlen = idpt, idlen

   prerr()
   mderpr("Illegal input character")
   prctxt()

   idpt, idlen := oidpt, oidlen
$)

AND ersnw() BE
/* Illegal name after warning.  */
$( prerr()
   mderpr("Illegal macro name after warning")
   prviz()
$)

AND macerr(n) BE
/* System error.  This should never (!) occur. */
$( prerr()
   mderpr("System error %N*N*N", n)
   IF debugging_ THEN
      // abort(n)                          // Get a backtrace, etc.
   mihalt()
$)

AND mcabrt() BE
/* Routine to clear up and print diagnostics after an error in a call of
an operation macro or insert.  The aborted construction is given a  null
value.  Control is returned to the main scanning loop.  */
$( opexit()
   prname(mtchpt, TRUE)
   mderpr(" aborted due to above error*N*N")
   longjump(scanlevel, bsnext)
$)

AND prctxt() BE
/* Routine to print the current text.  */
$( LET erbloc = VEC edbsz - 1           // Error block - area for saving the EDB

   fmove(edbsz, at_edb, erbloc)         // Save the current EDB
   mderpr("*N*Ndetected in*N")
   IF dbugsw = db_eval GOTO erop

prct2:
   mderpr("line %N of ", linect)

   $( SWITCHON dbugsw INTO
      $( CASE db_source:                // In source text
            mderpr("source text*N*N")
            fmove(edbsz, erbloc, at_edb)// Restore the EDB
            RETURN
   
         CASE db_oparg:
         CASE db_subarg:                // In argument
            mderpr("argument %N of ", argno)
            GOTO erop
   
         CASE db_delim:                 // In delimiter
            mderpr("delimiter %N of ", argno)
   
         erop:                          // In operation macro, argument or delimiter
            argpt := dbugpt
            dbugpt := dbugpt!11         // Stacked value of mtchpt
            IF dbugsw NE db_eval THEN
            $( prname(dbugpt, FALSE)
               mderpr(" evaluated in*N")
               UNLESS dbugsw = db_oparg ENDCASE
            $)
   
         CASE db_repl:                  // In macro
            prname(dbugpt, FALSE)
            mderpr(" with ")
            IF !argpt = 0 THEN mderpr("no ")
            mderpr("arguments")
            argno := 0
            dbugpt := argpt
   
            WHILE !argpt NE argno DO
            $( argno := argno + 1
               mderpr("*N%I3) ", argno)
               setpts(db_subarg)
               prid()
            $)
   
            mderpr("*Ncalled from*N")
            ENDCASE
   
         DEFAULT:
            macerr(16)
      $)
      fmove(edbsz, stakpt + 1, at_edb)
      IF mchlin = linect GOTO prct2
      IF !spt = '*N' THEN
      $( linect := linect - 1
         IF linect = mchlin GOTO prct2
      $)
      mderpr("lines %N to %N of ", mchlin, linect)
   $) REPEAT
$)

AND prenv() BE
/* Routine to print the version number of the machine-independent logic,
and  the names of all the constructions in the current environment.  The
built-in operation macro names are not printed.  */
$( mderpr("*N*N*NVersion BJB*N")

   FOR i = ty_stop TO ty_skip DO
   $( mderpr("*N%Ss are*N*N", prtype(i, TRUE))

      FOR j = hashpt TO hashpt + lhv - 1 DO
      $( LET ptr = j
         $( ptr := !ptr
            IF ptr = nullpt BREAK

            IF stffpt <= ptr < endpt THEN
                                        // Eliminate built-in macros
            $( LET type = setype(ptr + 1)

               IF type = i THEN
               $( IF (type NE ty_stop) & NOT ckvaly(ptr, type) LOOP
                  prlid(ptr + 1)
                  mderpr("*N")
               $)
            $)
         $) REPEAT
      $)
   $)
$)

AND prerr() BE
/*  Routine to introduce diagnostic output.  The count of errors (in S5)
is updated.  */
$( mderpr("*N*NError(s)*N")
   !at_s5 := !at_s5 + 1
$)

AND prid() BE
/*   Routine  to  print  the  atom  described  by  'idpt'  and  'idlen',
restricting length to 'texmax' characters.  */
$( LET sw_, pt = ?, ?

   /* Handle null atom */

   IF idlen = 0 THEN
   $( mderpr("(null)")
      RETURN
   $)

   /* Test if layout character */

   IF idlen = 1 THEN
   $( pt := lulayk(TRUE)
      IF pt NE 0 THEN                   // Print keyword for layout character
      $( idlen := pt!2
         idpt := pt + 3
         mderpr("(%Z)")
         RETURN
      $)
   $)

   /* Print ordinary text, restricting length if necessary */

   $( sw_ := FALSE
      IF idlen > texmax THEN
      $( sw_ := TRUE
         pt := idpt + idlen - htmax
         idlen := htmax
      $)
      mderpr("%Z")
      UNLESS sw_ BREAK
      idpt := pt
      mderpr("  ---  ")
   $) REPEAT
$)

AND prlid(ptr) BE
/* Routine to print a LID - parameter points at orlink.  */
$( idlen, idpt := ptr!1, ptr + 2
   ptr := idpt + idlen                  // Move past atom
   prid()
   IF !ptr = withmk LOOP                // ...A WITH B...
   IF !ptr NE wthsmk BREAK              // ...A WITHS B...
   mderpr(" ")
$) REPEAT

AND prmiss() BE
/* Routine to print names of unmatched constructions. */
$( LET chlink = ?

   IF nestlv NE 0 THEN
   $( lfpt := cllfpt
      $( mderpr("Delimiter ")
         $( prlid(delpt)
            chlink := !delpt
            IF chlink = endchn BREAK
            mderpr(" or ")
            delpt := delpt + chlink
         $) REPEAT
         mderpr(" of ")
         prname(mtchpt, FALSE)
         prnfnd(mchlin)
      $) REPEATWHILE decalv()
      skiplv := 0
      IF skval < 0 THEN
         skval := -skval - 1
   $)

   IF skval NE 0 THEN
   $( mderpr("Label %N referenced", skval)
      prnfnd(sklin)
      skval := 0
   $)
   prctxt()
$)

AND prname(ptr, first_) BE
/* Routine to print the type and name  of  a  construction  -  parameter
points at orlink.  */
$( mderpr("%S ", prtype(setype(ptr), first_))
   prlid(ptr)
$)

AND prnfnd(line) BE
/* Routine to print "not found" message, with line number.  */
   mderpr(" in line %N of current text not found*N", line)

AND prtype(type, first_) = VALOF
/*  Function  to  return  a  string  describing  a  particular  type  of
construction.  */
$( LET mes = VALOF SWITCHON type INTO
   $( CASE ty_stop  :  RESULTIS "stop"
      CASE ty_macro :  RESULTIS "macro"
      CASE ty_warn  :  RESULTIS "warning"
      CASE ty_insert:  RESULTIS "insert"
      CASE ty_skip  :  RESULTIS "skip"
      DEFAULT: macerr(17)
   $)
   TEST first_ THEN
   $( FOR i = 0 TO getbyte(mes, 0) DO             // Copy string
         putbyte(typvec, i, getbyte(mes, i))

      putbyte(typvec, 1, getbyte(mes, 1) - 'a' + 'A')
      RESULTIS typvec
   $)
   OR RESULTIS mes
$)

AND prviz() BE
/* Routine to print fuller information, after an illegal macro  name  or
argument has been detected.  */
$( mderpr(", viz *"")
   prid()
   mderpr("*"")
   prctxt()
$)

AND setype(ptr) = VALOF
/*  Function  to return the type of a construction - parameter points at
orlink.  */
$( ptr := ptr + 1

   $( ptr := !ptr + ptr + 1
      UNLESS (!ptr = withmk) \/ (!ptr = wthsmk) BREAK
      ptr := ptr + 1
   $) REPEAT

   IF !ptr = spcsmk THEN ptr := ptr + 1

   RESULTIS ptr!1
$)

 .
