
// Segment 6 - Table initialisation code

GET "ML1HDR"

STATIC $( names = 0 $)

LET init_tables() BE
/* Initialises the global hash table,  and  inserts  the  names  of  the
operation macros.  The secondary delimiter chain is set up and linked to
the operation macro entries.  */
$( names := (TABLE

   /* Table of operation macro names */

   0,endchn,5,'M','C','D','E','F',262,ty_macro,locmk,mc_def,
   0,endchn,6,'M','C','S','K','I','P',337,ty_macro,locmk,mc_skip,
   0,endchn,5,'M','C','I','N','S',325,ty_macro,locmk,mc_ins,
   0,endchn,6,'M','C','W','A','R','N',317,ty_macro,locmk,mc_warn,

   0,endchn,6,'M','C','D','E','F','G',211,ty_macro,opmk,mc_def,
   0,endchn,7,'M','C','S','K','I','P','G',285,ty_macro,opmk,mc_skip,
   0,endchn,6,'M','C','I','N','S','G',272,ty_macro,opmk,mc_ins,
   0,endchn,7,'M','C','W','A','R','N','G',263,ty_macro,opmk,mc_warn,

   0,endchn,7,'M','C','N','O','D','E','F',endchn,ty_macro,locmk,mc_nodef,
   0,endchn,8,'M','C','N','O','S','K','I','P',endchn,ty_macro,locmk,mc_noskip,
   0,endchn,7,'M','C','N','O','I','N','S',endchn,ty_macro,locmk,mc_noins,
   0,endchn,8,'M','C','N','O','W','A','R','N',endchn,ty_macro,locmk,mc_nowarn,

   0,endchn,5,'M','C','S','E','T',162,ty_macro,opmk,mc_set,
   0,endchn,4,'M','C','G','O',111,ty_macro,opmk,mc_go,
   0,endchn,7,'M','C','A','L','T','E','R',142,ty_macro,opmk,mc_alter,
   0,endchn,6,'M','C','N','O','T','E',155,ty_macro,opmk,mc_note,
   0,endchn,6,'M','C','P','V','A','R',142,ty_macro,opmk,mc_pvar,

   0,endchn,6,'M','C','L','E','N','G',wthsmk,1,'(',116,ty_macro,opmk,mc_leng,
   0,endchn,5,'M','C','S','U','B',wthsmk,1,'(',91,ty_macro,opmk,mc_sub,

   0,endchn,6,'M','C','S','T','O','P',98,ty_macro,locmk,mc_stop,

   /*  This  is the table of secondary delimiters (for operation macros)
   and keywords.  Each operation macro entry is linked to the  chain  of
   its  secondary  delimiters  (implicitly,  by  the  code for it).  The
   keywords are linked via separate global pointers.  The  entire  table
   has  a  separate  chain  running  through it, linking each individual
   entry; this allows the operation macro MC-ALTER to locate and  change
   a secondary delimiter or a keyword.  */

   8,
   8,4,'V','A','R','S',2,
   6,
   6,2,'A','S',81,
   8,
   endchn,4,'S','S','A','S',73,

   6,
   6,2,'I','F',12,
   10,
   65,6,'U','N','L','E','S','S',2,
   6,
   6,2,'G','E',51,
   6,
   6,2,'G','R',45,
   6,
   6,2,'E','N',39,
   6,
   6,2,'B','C',33,
   5,
   endchn,1,'=',28,

   6,
   endchn,2,'T','O',22,
   5,
   endchn,1,',',2,
   5,
   endchn,1,',',2,
   5,
   endchn,1,')',endchn,
   5,
   5,1,',',2,
   5,
   endchn,1,'*N',endchn,

   /* Chain of layout characters */

   9,
   86,5,'S','P','A','C','E',endchn,
   6,
   74,2,'N','L',endchn,
   7,
   69,3,'T','A','B',endchn,
   6,
   63,2,'S','L',endchn,
   10,
   54,6,'S','P','A','C','E','S',endchn,

   /* Node marker */

   5,
   endchn,1,'N',endchn,

   /* Keyword chain - subchain of delimiter chain */

   8,kw_with,4,'W','I','T','H',endchn,
   9,kw_withs,5,'W','I','T','H','S',endchn,
   7,kw_opt,3,'O','P','T',endchn,
   6,kw_or,2,'O','R',endchn,
   7,kw_all,3,'A','L','L',endchn,
   endchn,

   /* Dummy last entry */

   endchn,0,

   /* Machine-dependent tables */

   '*S','*N','*T',slrep,'*S'

      )

   init_ghshtb()
   init_delchn()
$)

AND init_ghshtb() BE
/* Initialises the hash table entries for the operation macros.  */
$( LET ptr = 0

   FOR i = 0 TO lhv - 1 DO              // Initialise heads of hash chains
      ghshtb!i := nullpt

   $( idlen := names!2
      idpt := names + 3

      ptr := mdfind()
      !names := !ptr
      !ptr := names

      names := idpt + idlen

      IF !names = wthsmk THEN           // Handle multi-atom names
         names := names + names!1 + 2

      names := names + 4                // Move to next entry
   $) REPEATUNTIL names!(-1) = mc_stop
$)


AND init_delchn() BE
/*  Routine  to  initialise addresses of items in master delimiter chain
and secondary delimiter chains.  */
$( LET do_entry(atmdc, atsdc) BE
   $( IF atmdc NE 0 THEN !atmdc := names
      IF atsdc NE 0 THEN !atsdc := names + 1

      names := names + !names
   $)

   do_entry(@delchn, 0)                 // VARS
   do_entry(0, @das)                    // AS
   do_entry(0, 0)                       // SSAS
   do_entry(0, @dif)                    // IF
   do_entry(0, 0)                       // UNLESS
   do_entry(0, @dge)                    // GE
   do_entry(0, 0)                       // GR
   do_entry(0, 0)                       // EN
   do_entry(0, 0)                       // BC
   do_entry(0, 0)                       // EQ
   do_entry(0, 0)                       // TO
   do_entry(0, 0)                       // ,
   do_entry(0, 0)                       // ,
   do_entry(0, 0)                       // )
   do_entry(0, 0)                       // ,
   do_entry(0, 0)                       // <newline>
   do_entry(@laychn, 0)                 // SPACE
   do_entry(0, 0)                       // NL
   do_entry(0, 0)                       // TAB
   do_entry(0, 0)                       // SL
   do_entry(@kspacs, 0)                 // SPACES
   do_entry(@knrep, 0)                  // N
   do_entry(@keychn, 0)                 // WITH
   do_entry(0, 0)                       // WITHS
   do_entry(0, 0)                       // OPT
   do_entry(0, 0)                       // OR
   do_entry(0, 0)                       // ALL

   spcsrp := kspacs + kspacs!1 + 1
   knrep := knrep + 3

   IF debugging_ THEN
      UNLESS !names = endchn DO macerr(18)
$)

 .
